@@ -1,11 +1,12 @@
 
 ansi_string <- function(x) {
   if (!is.character(x)) x <- as.character(x)
-  class(x) <- unique(c("ansi_string", class(x)))
+  x <- enc2utf8(x)
+  class(x) <- unique(c("cli_ansi_string", "ansi_string", class(x), "character"))
   x
 }
 
-#' Perl comparible regular expression that matches ANSI escape
+#' Perl compatible regular expression that matches ANSI escape
 #' sequences
 #'
 #' Don't forget to use `perl = TRUE` when using this with [grepl()] and
@@ -30,6 +31,8 @@ ansi_regex <- function() {
 #'
 #' @param string The string to check. It can also be a character
 #'   vector.
+#' @param sgr Whether to look for SGR (styling) control sequences.
+#' @param csi Whether to look for non-SGR control sequences.
 #' @return Logical vector, `TRUE` for the strings that have some
 #'   ANSI styling.
 #'
@@ -40,75 +43,53 @@ ansi_regex <- function() {
 #' ansi_has_any("foobar")
 #' ansi_has_any(col_red("foobar"))
 
-ansi_has_any <- function(string) {
-  grepl(ansi_regex(), string, perl = TRUE)
+ansi_has_any <- function(string, sgr = TRUE, csi = TRUE) {
+  if (!is.character(string)) string <- as.character(string)
+  string <- enc2utf8(string)
+  stopifnot(
+    is_flag(sgr),
+    is_flag(csi)
+  )
+  .Call(clic_ansi_has_any, string, sgr, csi)
 }
 
 #' Remove ANSI escape sequences from a string
 #'
-#' The input may be of class `ansi_string` class, this is also dropped
+#' The input may be of class `cli_ansi_string` class, this is also dropped
 #' from the result.
 #'
 #' @param string The input string.
-#' @return The cleaned up string.
+#' @param sgr Whether to remove for SGR (styling) control sequences.
+#' @param csi Whether to remove for non-SGR control sequences.
+#' @return The cleaned up string. Note that `ansi_strip()` always drops
+#' the `cli_ansi_string` class, even if `sgr` and sci` are `FALSE`.
 #'
 #' @family low level ANSI functions
 #' @export
 #' @examples
 #' ansi_strip(col_red("foobar")) == "foobar"
 
-ansi_strip <- function(string) {
-  clean <- gsub(ansi_regex(), "", string, perl = TRUE)
-  class(clean) <- setdiff(class(clean), "ansi_string")
+ansi_strip <- function(string, sgr = TRUE, csi = TRUE) {
+  if (!is.character(string)) string <- as.character(string)
+  string <- enc2utf8(string)
+  stopifnot(
+    is_flag(sgr),
+    is_flag(csi)
+  )
+  clean <- .Call(clic_ansi_strip, string, sgr, csi)
+  class(clean) <- setdiff(class(clean), c("cli_ansi_string", "ansi_string"))
   clean
 }
 
-
-## Create a mapping between the string and its style-less version.
-## This is useful to work with the colored string.
-
-#' @importFrom utils tail
-
-map_to_ansi <- function(x, text = NULL) {
-
-  if (is.null(text)) {
-    text <- non_matching(re_table(ansi_regex(), x), x, empty=TRUE)
-  }
-
-  map <- lapply(
-    text,
-    function(text) {
-      cbind(
-        pos = cumsum(c(1, text[, "length"], Inf)),
-        offset = c(text[, "start"] - 1, tail(text[, "end"], 1), NA)
-      )
-    })
-
-  function(pos) {
-    pos <- rep(pos, length.out = length(map))
-    mapply(pos, map, FUN = function(pos, table) {
-      if (pos < 1) {
-        pos
-      } else {
-        slot <- which(pos < table[, "pos"])[1] - 1
-        table[slot, "offset"] + pos - table[slot, "pos"] + 1
-      }
-    })
-  }
-}
-
 #' Count number of characters in an ANSI colored string
 #'
-#' This is a color-aware counterpart of [base::nchar()],
-#' which does not do well, since it also counts the ANSI control
-#' characters.
+#' This is a color-aware counterpart of [utf8_nchar()]. By default it
+#' counts Unicode grapheme clusters, instead of code points.
 #'
-#' @param x Character vector, potentially ANSO styled, or a vector to be
-#'   coarced to character.
-#' @param type Whether to count characters, bytes, or calculate the
-#'   display width of the string. Passed to [base::nchar()].
-#' @param ... Additional arguments, passed on to [base::nchar()]
-#'   after removing ANSI escape sequences.
+#' @param x Character vector, potentially ANSI styled, or a vector to be
+#'   coerced to character. If it converted to UTF-8.
+#' @param type Whether to count graphemes (characters), code points,
+#'   bytes, or calculate the display width of the string. 
 #' @return Numeric vector, the length of the strings in the character
 #'   vector.
 #'
@@ -126,14 +107,15 @@ map_to_ansi <- function(x, text = NULL) {
 #' ansi_nchar(str)
 #' nchar(ansi_strip(str))
 
-ansi_nchar <- function(x, type = c("chars", "bytes", "width"), ...) {
+ansi_nchar <- function(x,
+                       type = c("chars", "bytes", "width", "graphemes",
+                                "codepoints")) {
   type <- match.arg(type)
-  if (type == "width") x <- unicode_pre(x)
-  ansi_nchar_bad(x, type = type, ...)
-}
-
-ansi_nchar_bad <- function(x, ...) {
-  base::nchar(ansi_strip(x), ...)
+  if (type == "chars") type <- "graphemes"
+  type <- match(type, c("graphemes", "bytes", "width", "codepoints"))
+  if (!is.character(x)) x <- as.character(x)
+  x <- enc2utf8(x)
+  .Call(clic_ansi_nchar, x, type)
 }
 
 #' Substring(s) of an ANSI colored string
@@ -144,7 +126,7 @@ ansi_nchar_bad <- function(x, ...) {
 #' calculating the positions within the string.
 #'
 #' @param x Character vector, potentially ANSI styled, or a vector to
-#'   coarced to character.
+#'   coerced to character.
 #' @param start Starting index or indices, recycled to match the length
 #'   of `x`.
 #' @param stop Ending index or indices, recycled to match the length
@@ -191,26 +173,10 @@ ansi_substr <- function(x, start, stop) {
   if (anyNA(start) || anyNA(stop)) {
     stop("non-numeric substring arguments not supported")
   }
-  ansi <- re_table(ansi_regex(), x)
-  text <- non_matching(ansi, x, empty=TRUE)
-  mapper <- map_to_ansi(x, text = text)
-  ansi_substr_internal(x, mapper, start, stop)
-}
-
-ansi_substr_internal <- function(x, mapper, start, stop) {
-  nstart <- mapper(start)
-  nstop  <- mapper(stop)
-
-  bef <- base::substr(x, 1, nstart - 1)
-  aft <- base::substr(x, nstop + 1, base::nchar(x))
-  ansi_bef <- vapply(regmatches(bef, gregexpr(ansi_regex(), bef)),
-                     paste, collapse = "", FUN.VALUE = "")
-  ansi_aft <- vapply(regmatches(aft, gregexpr(ansi_regex(), aft)),
-                     paste, collapse = "", FUN.VALUE = "")
-
-  ansi_string(
-    paste(sep = "", ansi_bef, base::substr(x, nstart, nstop), ansi_aft)
-  )
+  x <- enc2utf8(x)
+  start <- rep_len(start, length(x))
+  stop <- rep_len(stop, length(x))
+  .Call(clic_ansi_substr, x, start, stop)
 }
 
 #' Substring(s) of an ANSI colored string
@@ -221,7 +187,7 @@ ansi_substr_internal <- function(x, mapper, start, stop) {
 #' calculating the positions within the string.
 #'
 #' @param text Character vector, potentially ANSI styled, or a vector to
-#'   coarced to character. It is recycled to the longest of `first`
+#'   coerced to character. It is recycled to the longest of `first`
 #'   and `last`.
 #' @param first Starting index or indices, recycled to match the length
 #'   of `x`.
@@ -262,7 +228,10 @@ ansi_substring <- function(text, first, last = 1000000L) {
   if (!is.character(text)) text <- as.character(text)
   n <- max(lt <- length(text), length(first), length(last))
   if (lt && lt < n) text <- rep_len(text, length.out = n)
-  ansi_substr(text, as.integer(first), as.integer(last))
+  text <- enc2utf8(text)
+  first <- rep_len(as.integer(first), n)
+  last <- rep_len(as.integer(last), n)
+  .Call(clic_ansi_substr, text, first, last)
 }
 
 
@@ -273,19 +242,18 @@ ansi_substring <- function(text, first, last = 1000000L) {
 #' substrings.
 #'
 #' @param x Character vector, potentially ANSI styled, or a vector to
-#'   coarced to character.
+#'   coerced to character.
 #' @param split Character vector of length 1 (or object which can be coerced to
 #'   such) containing regular expression(s) (unless `fixed = TRUE`) to use
 #'   for splitting.  If empty matches occur, in particular if `split` has
 #'   zero characters, `x` is split into single characters.
 #' @param ... Extra arguments are passed to `base::strsplit()`.
-#' @return A list of the same length as `x`, the \eqn{i}-th element of
+#' @return A list of the same length as `x`, the `i`-th element of
 #'   which contains the vector of splits of `x[i]`. ANSI styles are
 #'   retained.
 #'
 #' @family ANSI string operations
 #' @export
-#' @importFrom utils head
 #' @examples
 #' str <- paste0(
 #'   col_red("I am red---"),
@@ -307,6 +275,8 @@ ansi_strsplit <- function(x, split, ...) {
   split <- try(as.character(split), silent=TRUE)
   if(inherits(split, "try-error") || !is.character(split) || length(split) > 1L)
     stop("`split` must be character of length <= 1, or must coerce to that")
+  if (!is.character(x)) x <- as.character(x)
+  x <- enc2utf8(x)
   if(!length(split)) split <- ""
   plain <- ansi_strip(x)
   splits <- re_table(split, plain, ...)
@@ -320,11 +290,11 @@ ansi_strsplit <- function(x, split, ...) {
     function(i) {
       y <- chunks[[i]]
       # empty split means drop empty first match
-      if(nrow(y) && !nzchar(split.r[[i]]) && !head(y, 1L)[, "length"]) {
+      if(nrow(y) && !nzchar(split.r[[i]]) && !utils::head(y, 1L)[, "length"]) {
         y <- y[-1L, , drop=FALSE]
       }
       # drop empty last matches
-      if(nrow(y) && !tail(y, 1L)[, "length"]) y[-nrow(y), , drop=FALSE] else y
+      if(nrow(y) && !utils::tail(y, 1L)[, "length"]) y[-nrow(y), , drop=FALSE] else y
     }
   )
   zero.chunks <- !vapply(chunks, nrow, integer(1L))
@@ -341,6 +311,35 @@ ansi_strsplit <- function(x, split, ...) {
 
 #' Align an ANSI colored string
 #'
+#' @details
+#'
+#' ```{asciicast ansi-align}
+#' str <- c(
+#'   col_red("This is red"),
+#'   style_bold("This is bold")
+#' )
+#' astr <- ansi_align(str, width = 30)
+#' boxx(astr)
+#' ```
+#'
+#' ```{asciicast ansi-align-center}
+#' str <- c(
+#'   col_red("This is red"),
+#'   style_bold("This is bold")
+#' )
+#' astr <- ansi_align(str, align = "center", width = 30)
+#' boxx(astr)
+#' ```
+#'
+#' ```{asciicast ansi-align-right}
+#' str <- c(
+#'   col_red("This is red"),
+#'   style_bold("This is bold")
+#' )
+#' astr <- ansi_align(str, align = "right", width = 30)
+#' boxx(astr)
+#' ```
+#'
 #' @param text The character vector to align.
 #' @param width Width of the field to align in.
 #' @param align Whether to align `"left"`, `"center"` or `"right"`.
@@ -349,16 +348,15 @@ ansi_strsplit <- function(x, split, ...) {
 #'
 #' @family ANSI string operations
 #' @export
-#' @examples
-#' ansi_align(col_red("foobar"), 20, "left")
-#' ansi_align(col_red("foobar"), 20, "center")
-#' ansi_align(col_red("foobar"), 20, "right")
+
+# TODO: show wide Unicode charadcters, once they work in asciicast
 
 ansi_align <- function(text, width = console_width(),
                       align = c("left", "center", "right"),
                       type = "width") {
 
   align <- match.arg(align)
+  text <- enc2utf8(text)
   nc <- ansi_nchar(text, type = type)
 
   if (!length(text)) return(ansi_string(text))
@@ -425,6 +423,7 @@ ansi_trimws <- function(x, which = c("both", "left", "right")) {
 
   if (!is.character(x)) x <- as.character(x)
   which <- match.arg(which)
+  x <- enc2utf8(x)
   if (!length(x)) return(ansi_string(x))
 
   sl <- 0L
@@ -444,7 +443,8 @@ ansi_trimws <- function(x, which = c("both", "left", "right")) {
   }
 
   if (any(sl > 0L | rl > 0L)) {
-    x <- ansi_substr(x, 1 + sl, ansi_nchar(x) - rl)
+    start <- rep_len(1L + sl, length(x))
+    x <- .Call(clic_ansi_substr, x, start, ansi_nchar(x) - rl)
   }
 
   ansi_string(x)
@@ -460,7 +460,7 @@ ansi_trimws <- function(x, which = c("both", "left", "right")) {
 #' @param indent Indentation of the first line of each paragraph.
 #' @param exdent Indentation of the subsequent lines of each paragraph.
 #' @param simplify Whether to return all wrapped strings in a single
-#'   charcter vector, or wrap each element of `x` independently and return
+#'   character vector, or wrap each element of `x` independently and return
 #'   a list.
 #' @return If `simplify` is `FALSE`, then a list of character vectors,
 #'   each an ANSI string. Otherwise a single ANSI string vector.
@@ -481,6 +481,7 @@ ansi_strwrap <- function(x, width = console_width(), indent = 0,
                          exdent = 0, simplify = TRUE) {
 
   if (!is.character(x)) x <- as.character(x)
+  x <- enc2utf8(x)
   if (length(x) == 0) {
     return(ansi_string(x))
   }
@@ -494,6 +495,21 @@ ansi_strwrap <- function(x, width = console_width(), indent = 0,
   # Workaround for bad Unicode width
   x <- unicode_pre(x)
 
+  # Form feeds are forced line breaks
+  # R 4.2 removes the \f after <https://github.com/wch/r-source/commit/101b142d04dd5456a2039d54de9483240bcc1512>
+  # se we need to put in a random marker instead
+  mark <- "yShtnpteEk"
+  smark <- paste0("\n\n", mark, "\n\n")
+  x <- gsub("\f", smark, x, fixed = TRUE, useBytes = TRUE)
+  fix_ff <- function(x) {
+    rem <- which(ansi_strip(x) == mark)
+    if (length(rem)) {
+      x[-c(rem - 1, rem, rem + 1)]
+    } else {
+      x
+    }
+  }
+
   # First we need to remove the multiple spaces, to make it easier to
   # map the strings later on. We do this per paragraph, to keep paragraphs.
   pars <- strsplit(x, "\n[ \t\n]*\n", perl = TRUE)
@@ -517,7 +533,7 @@ ansi_strwrap <- function(x, width = console_width(), indent = 0,
 
   xs <- ansi_strip(xx)
   xw0 <- base::strwrap(xs, width = width, indent = indent, exdent = exdent)
-  if (xs == xx) return(ansi_string(unicode_post(xw0)))
+  if (xs == xx) return(ansi_string(unicode_post(fix_ff(xw0))))
 
   xw <- trimws(xw0, "left")
   indent <- nchar(xw0) - nchar(xw)
@@ -533,7 +549,10 @@ ansi_strwrap <- function(x, width = console_width(), indent = 0,
   while (xsidx <= xslen) {
     xsc <- substr(xs, xsidx, xsidx)
     xwc <- substr(xw[xwidx[1]], xwidx[2], xwidx[2])
-    if (xsc == xwc) {
+    if (is.na(xwc)) {
+      # colored trailing white space in input?
+      xsidx <- xsidx + 1L
+    } else if (xsc == xwc) {
       xsidx <- xsidx + 1L
       xwidx[2] <- xwidx[2] + 1L
     } else if (xsc %in% c(" ", "\n", "\t")) {
@@ -554,24 +573,20 @@ ansi_strwrap <- function(x, width = console_width(), indent = 0,
   }
   splits <- c(splits, xsidx)
 
-  ansi <- re_table(ansi_regex(), xx)
-  text <- non_matching(ansi, xx, empty=TRUE)
-  mapper <- map_to_ansi(xx, text = text)
-
   wrp <- vcapply(seq_along(splits[-1]), function(i) {
     from <- splits[i]
     to <- splits[i + 1L] - 1L
     while (from %in% drop) from <- from + 1L
-    ansi_substr_internal(xx, mapper, from, to)
+    .Call(clic_ansi_substr, xx, from, to)
   })
 
   indent <- strrep(" ", indent)
-  ansi_string(unicode_post(paste0(indent, wrp)))
+  ansi_string(unicode_post(fix_ff(paste0(indent, wrp))))
 }
 
 #' Truncate an ANSI string
 #'
-#' This function is similar to [base::strtrim()], but works correcntly with
+#' This function is similar to [base::strtrim()], but works correctly with
 #' ANSI styled strings. It also adds `...` (or the corresponding Unicode
 #' character if Unicode characters are allowed) to the end of truncated
 #' strings.
@@ -590,22 +605,31 @@ ansi_strwrap <- function(x, width = console_width(), indent = 0,
 ansi_strtrim <- function(x, width = console_width(),
                          ellipsis = symbol$ellipsis) {
 
+  x <- enc2utf8(x)
+
   # Unicode width notes. We have nothing to fix here, because we'll just
   # use ansi_substr() and ansi_nchar(), which work correctly with wide
   # characters.
 
   # First we cut according to _characters_. This might be too wide if we
   # have wide characters.
-  xt <- ansi_substr(x, 1, width)
+  lx <- length(x)
+  xt <- .Call(clic_ansi_substr, x, rep(1L, lx), rep(as.integer(width), lx))
   tw <- ansi_nchar(ellipsis, "width")
 
-  # If there was a cut, or xt is too wise (using _width_!), that's bad
+  # If there was a cut, or xt is too wide (using _width_!), that's bad
   # We keep the initial bad ones, these are the ones that need an ellipsis.
   # Then we keep chopping off single characters from the too wide ones,
   # until they are narrow enough.
-  bad0 <- bad <- xt != x | ansi_nchar(xt, "width") > width
+  bad0 <- bad <- !is.na(x) &
+    (ansi_strip(xt) != ansi_strip(x) | ansi_nchar(xt, "width") > width)
   while (any(bad)) {
-    xt[bad] <- ansi_substr(xt[bad], 1, ansi_nchar(xt[bad]) - 1L)
+    xt[bad] <- .Call(
+      clic_ansi_substr,
+      xt[bad],
+      rep(1L, sum(bad)),
+      ansi_nchar(xt[bad]) - 1L
+    )
     bad <- ansi_nchar(xt, "width") > width - tw
   }
 
@@ -621,6 +645,18 @@ ansi_strtrim <- function(x, width = console_width(),
 #' If a string does not fit into the specified `width`, it will be
 #' truncated using [ansi_strtrim()].
 #'
+#' ```{asciicast ansi-column}
+#' fmt <- ansi_columns(
+#'   paste(col_red("foo"), 1:10),
+#'   width = 50,
+#'   fill = "rows",
+#'   max_cols=10,
+#'   align = "center",
+#'   sep = "   "
+#' )
+#' boxx(fmt, padding = c(0,1,0,1), header = col_cyan("Columns"))
+#' ```
+#'
 #' @param text Character vector to format. Each element will formatted
 #'   as a cell of a table.
 #' @param width Width of the screen.
@@ -636,18 +672,6 @@ ansi_strtrim <- function(x, width = console_width(),
 #'
 #' @family ANSI string operations
 #' @export
-#' @examples
-#' fmt <- ansi_columns(
-#'   paste(col_red("foo"), 1:10),
-#'   width = 50,
-#'   fill = "rows",
-#'   max_cols=10,
-#'   align = "center",
-#'   sep = "   "
-#' )
-#' fmt
-#' ansi_nchar(fmt, type = "width")
-#' boxx(fmt, padding = c(0,1,0,1), header = col_green("foobar"))
 
 ansi_columns <- function(text, width = console_width(), sep = " ",
                          fill = c("rows", "cols"), max_cols = 4,
@@ -657,6 +681,8 @@ ansi_columns <- function(text, width = console_width(), sep = " ",
   fill <- match.arg(fill)
   align <- match.arg(align)
 
+  text <- enc2utf8(text)
+
   if (length(text) == 0) return(ansi_string(text))
 
   swdh <- ansi_nchar(sep, type = "width")
@@ -685,3 +711,222 @@ ansi_columns <- function(text, width = console_width(), sep = " ",
   clp <- apply(tm, 1, paste0, collapse = "")
   ansi_string(clp)
 }
+
+#' ANSI character translation and case folding
+#'
+#' There functions are similar to [toupper()], [tolower()] and
+#' [chartr()], but they keep the ANSI colors of the string.
+#'
+#' @inheritParams base::chartr
+#' @param x Input string. May have ANSI colors and styles.
+#' @return Character vector of the same length as `x`, containing
+#'   the translated strings. ANSI styles are retained.
+#'
+#' @family ANSI string operations
+#' @export
+#' @examples
+#' ansi_toupper(col_red("Uppercase"))
+#'
+#' ansi_tolower(col_red("LowerCase"))
+#'
+#' x <- paste0(col_green("MiXeD"), col_red(" cAsE 123"))
+#' ansi_chartr("iXs", "why", x)
+
+ansi_toupper <- function(x) {
+  ansi_convert(x, toupper)
+}
+
+#' @family ANSI string operations
+#' @export
+#' @rdname ansi_toupper
+
+ansi_tolower <- function(x) {
+  ansi_convert(x, tolower)
+}
+
+#' @family ANSI string operations
+#' @export
+#' @rdname ansi_toupper
+
+ansi_chartr <- function(old, new, x) {
+  ansi_convert(x, chartr, old, new)
+}
+
+ansi_convert <- function(x, converter, ...) {
+  x <- enc2utf8(x)
+  ansi <- re_table(ansi_regex(), x)
+  text <- non_matching(ansi, x, empty=TRUE)
+  out <- mapply(x, text, USE.NAMES = FALSE, FUN = function(x1, t1) {
+    t1 <- t1[t1[,1] <= t1[,2], , drop = FALSE]
+    for (i in seq_len(nrow(t1))) {
+      substring(x1, t1[i, 1], t1[i, 2]) <-
+        converter(x = substring(x1, t1[i, 1], t1[i, 2]), ...)
+    }
+    x1
+  })
+
+  ansi_string(out)
+}
+
+#' Simplify ANSI styling tags
+#'
+#' It creates an equivalent, but possibly shorter ANSI styled string, by
+#' removing duplicate and empty tags.
+#'
+#' @param x Input string
+#' @param csi What to do with non-SGR ANSI sequences, either `"keep"`,
+#'   or `"drop"` them.
+#' @return Simplified `cli_ansi_string` vector.
+#'
+#' @export
+
+ansi_simplify <- function(x, csi = c("keep", "drop")) {
+  if (!is.character(x)) x <- as.character(x)
+  csi <- match.arg(csi)
+  x <- enc2utf8(x)
+  .Call(clic_ansi_simplify, x, csi == "keep")
+}
+
+#' Convert ANSI styled text to HTML
+#'
+#' @param x Input character vector.
+#' @param escape_reserved Whether to escape characters that are reserved
+#'   in HTML (`&`, `<` and `>`).
+#' @param csi What to do with non-SGR ANSI sequences, either `"keep"`,
+#'   or `"drop"` them.
+#' @return Character vector of HTML.
+#'
+#' @family ANSI to HTML conversion
+#' @export
+#' @examplesIf cli:::has_packages(c("htmltools", "withr"))
+#' ## Syntax highlight the source code of an R function with ANSI tags,
+#' ## and export it to a HTML file.
+#' code <- withr::with_options(
+#'   list(ansi.num_colors = 256),
+#'   code_highlight(format(ansi_html))
+#' )
+#' hcode <- paste(ansi_html(code), collapse = "\n")
+#' css <- paste(format(ansi_html_style()), collapse=  "\n")
+#' page <- htmltools::tagList(
+#'   htmltools::tags$head(htmltools::tags$style(css)),
+#'   htmltools::tags$pre(htmltools::HTML(hcode))
+#' )
+#'
+#' if (interactive()) htmltools::html_print(page)
+
+ansi_html <- function(x, escape_reserved = TRUE, csi = c("drop", "keep")) {
+  if (!is.character(x)) x <- as.character(x)
+  csi <- match.arg(csi)
+  x <- enc2utf8(x)
+  if (escape_reserved) {
+    x <- gsub("&", "&amp;", x, fixed = TRUE, useBytes = TRUE)
+    x <- gsub("<", "&lt;",  x, fixed = TRUE, useBytes = TRUE)
+    x <- gsub(">", "&gt;",  x, fixed = TRUE, useBytes = TRUE)
+  }
+  .Call(clic_ansi_html, x, csi == "keep")
+}
+
+#' CSS styles for the output of `ansi_html()`
+#'
+#'
+#'
+#' @param colors Whether or not to include colors. `FALSE` will not include
+#'   colors, `TRUE` or `8` will include eight colors (plus their bright
+#'   variants), `256` will include 256 colors.
+#' @param palette Character scalar, palette to use for the first eight colors
+#'   plus their bright variants. Terminals define these colors differently,
+#'   and cli includes a couple of examples. Sources of palettes:
+#'   * https://en.wikipedia.org/wiki/ANSI_escape_code#3-bit_and_4-bit
+#'   * iTerm2 builtin palettes
+#'   * <https://github.com/sindresorhus/iterm2-snazzy>
+#' @return Named list of CSS declaration blocks, where the names are
+#'   CSS selectors. It has a `format()` and `print()` methods, which you
+#'   can use to write the output to a CSS or HTML file.
+#'
+#' @family ANSI to HTML conversion
+#' @export
+#' @examples
+#' ansi_html_style(colors = FALSE)
+#' ansi_html_style(colors = 8, palette = "iterm-snazzy")
+
+ansi_html_style <- function(colors = TRUE, palette = NULL) {
+  if (is.character(palette)) {
+    palette <- match.arg(palette)
+    palette <- as.list(ansi_palettes[palette, ])
+  }
+
+  stopifnot(
+    isTRUE(colors) || identical(colors, FALSE) ||
+      (is_count(colors) && colors %in% c(8,256)),
+    is_string(palette) || is.list(palette) && length(palette) == 16
+  )
+
+  ret <- list(
+    ".ansi-bold"       = "{ font-weight: bold;             }",
+    # .ansi-faint ???
+    ".ansi-italic"     = "{ font-style: italic;            }",
+    ".ansi-underline"  = "{ text-decoration: underline;    }",
+    ".ansi-blink"      = "{ text-decoration: blink;        }",
+    # .ansi-inverse ???
+    ".ansi-hide"       = "{ visibility: hidden;            }",
+    ".ansi-crossedout" = "{ text-decoration: line-through; }"
+  )
+
+  if (!identical(colors, FALSE)) {
+    fg <- structure(
+      names = paste0(".ansi-color-", 0:15),
+      paste0("{ color: ", palette, " }")
+    )
+    bg <- structure(
+      names = paste0(".ansi-bg-color-", 0:15),
+      paste0("{ background-color: ", palette, " }")
+    )
+    ret <- c(ret, fg, bg)
+  }
+
+  if (isTRUE(colors) || colors == 256) {
+    grid <- expand.grid(r = 0:5, g = 0:5, b = 0:5)
+    num <- 16 + 36 * grid$r + 6 * grid$g + grid$b
+    cols <- grDevices::rgb(grid$r, grid$g, grid$b, maxColorValue = 5)
+    fg256 <- structure(
+      names = paste0(".ansi-color-", num),
+      paste0("{ color: ", tolower(cols), " }")
+    )
+    bg256 <- structure(
+      names = paste0(".ansi-bg-color-", num),
+      paste0("{ background-color: ", tolower(cols), " }")
+    )
+    gr <- seq(1, 24)
+    grcols <- grDevices::rgb(gr, gr, gr, maxColorValue = 25)
+    fggrey <- structure(
+      names = paste0(".ansi-color-", 232:255),
+      paste0("{ color: ", tolower(grcols), " }")
+    )
+    bggrey <- structure(
+      names = paste0(".ansi-bg-color-", 232:255),
+      paste0("{ background-color: ", tolower(grcols), " }")
+    )
+    ret <- c(ret, fg256, fggrey, bg256, bggrey)
+  }
+
+  class(ret) <- "cli_ansi_html_style"
+  ret
+}
+
+# This avoids duplication, but messes up the source ref of the function...
+formals(ansi_html_style)$palette <- c("vscode", setdiff(rownames(ansi_palettes), "vscode"))
+attr(body(ansi_html_style), "srcref") <- NULL
+attr(body(ansi_html_style), "wholeSrcref") <- NULL
+attr(body(ansi_html_style), "srcfile") <- NULL
+
+#' @export
+
+format.cli_ansi_html_style <- function(x, ...) {
+  paste0(format(names(x)), " ", x)
+}
+
+#' @export
+
+print.cli_ansi_html_style <- function(x, ...) {
+  cat(format(x, ...), sep = "\n")
+}
