@@ -1,5 +1,5 @@
 
-#' CLI pluralization
+#' About cli pluralization
 #'
 #' @name pluralization
 #' @family pluralization
@@ -183,6 +183,6 @@ pluralize <- function(..., .envir = parent.frame(),
     }
   }
 
-  raw <- glue::glue(..., .envir = .envir, .transformer = tf)
+  raw <- glue::glue(..., .envir = .envir, .transformer = tf, .comment = "")
   post_process_plurals(raw, values)
 }
