@@ -13,7 +13,7 @@
 #'   `NA` if the style is not used.
 #'
 #' @export
-#' @seealso themes
+#' @seealso [themes]
 
 cli_list_themes <- function() {
   app <- default_app() %||% start_app()
@@ -42,23 +42,60 @@ clii_remove_theme <- function(app, id) {
 #' This theme is always active, and it is at the bottom of the theme
 #' stack. See [themes].
 #'
+#' # Showcase
+#'
+#' ```{asciicast builtin-theme}
+#' cli_h1("Heading 1")
+#' cli_h2("Heading 2")
+#' cli_h3("Heading 3")
+#'
+#' cli_par()
+#' cli_alert_danger("Danger alert")
+#' cli_alert_warning("Warning alert")
+#' cli_alert_info("Info alert")
+#' cli_alert_success("Success alert")
+#' cli_alert("Alert for starting a process or computation",
+#'   class = "alert-start")
+#' cli_end()
+#'
+#' cli_text("Packages and versions: {.pkg cli} {.version 1.0.0}.")
+#' cli_text("Time intervals: {.timestamp 3.4s}")
+#'
+#' cli_text("{.emph Emphasis} and  {.strong strong emphasis}")
+#'
+#' cli_text("This is a piece of code: {.code sum(x) / length(x)}")
+#' cli_text("Function names: {.fn cli::simple_theme}")
+#'
+#' cli_text("Files: {.file /usr/bin/env}")
+#' cli_text("URLs: {.url https://r-project.org}")
+#'
+#' cli_h2("Longer code chunk")
+#' cli_par(class = "code R")
+#' cli_verbatim(
+#'   '# window functions are useful for grouped mutates',
+#'   'mtcars %>%',
+#'   '  group_by(cyl) %>%',
+#'   '  mutate(rank = min_rank(desc(mpg)))')
+#' ```
+#'
 #' @seealso [themes], [simple_theme()].
 #' @return A named list, a CLI theme.
 #'
-#' @param dark Whether to use a dark theme. The `cli_theme_dark` option
+#' @param dark Whether to use a dark theme. The `cli.theme_dark` option
 #'   can be used to request a dark theme explicitly. If this is not set,
 #'   or set to `"auto"`, then cli tries to detect a dark theme, this
 #'   works in recent RStudio versions and in iTerm on macOS.
 #' @export
 
-builtin_theme <- function(dark = getOption("cli_theme_dark", "auto")) {
+builtin_theme <- function(dark = getOption("cli.theme_dark", "auto")) {
 
   dark <- detect_dark_theme(dark)
 
   list(
     body = list(
       "class-map" = list(
-        fs_path = "file"
+        fs_path = "file",
+        "cli-progress-bar" = "progress-bar"
       )
     ),
 
@@ -93,26 +130,54 @@ builtin_theme <- function(dark = getOption("cli_theme_dark", "auto")) {
       before = function() paste0(col_cyan(symbol$info), " ")
     ),
 
-    par = list("margin-top" = 0, "margin-bottom" = 1),
-    li = list("padding-left" = 2),
-    ul = list(
-      "list-style-type" = function() symbol$bullet,
-      "padding-left" = 0
+    ".bullets .bullet-empty" = list(),
+    ".bullets .bullet-space" = list("margin-left" = 2),
+    ".bullets .bullet-v" = list(
+      "text-exdent" = 2,
+      before = function(x) paste0(col_green(symbol$tick), " ")
     ),
-    "ul ul" = list(
-      "list-style-type" = function() symbol$circle,
-      "padding-left" = 2
+    ".bullets .bullet-x" = list(
+      "text-exdent" = 2,
+      before = function(x) paste0(col_red(symbol$cross), " ")
+    ),
+    ".bullets .bullet-!" = list(
+      "text-exdent" = 2,
+      before = function(x) paste0(col_yellow("!"), " ")
+    ),
+    ".bullets .bullet-i" = list(
+      "text-exdent" = 2,
+      before = function(x) paste0(col_cyan(symbol$info), " ")
+    ),
+    ".bullets .bullet-*" = list(
+      "text-exdent" = 2,
+      before = function(x) paste0(col_cyan(symbol$bullet), " ")
+    ),
+    ".bullets .bullet->" = list(
+      "text-exdent" = 2,
+      before = function(x) paste0(symbol$arrow_right, " ")
+    ),
+    ".bullets .bullet-1" = list(
     ),
-    "ul ul ul" = list("list-style-type" = function() symbol$line),
 
-    "ul ul" = list("padding-left" = 2),
-    "ul dl" = list("padding-left" = 2),
-    "ol ol" = list("padding-left" = 2),
-    "ol ul" = list("padding-left" = 2),
-    "ol dl" = list("padding-left" = 2),
-    "dl ol" = list("padding-left" = 2),
-    "dl ul" = list("padding-left" = 2),
-    "dl dl" = list("padding-left" = 2),
+    par = list("margin-top" = 0, "margin-bottom" = 1),
+    ul = list(
+      "list-style-type" = function() symbol$bullet
+    ),
+
+    # these are tags in HTML, but in cli they are inline
+    span.dt = list(after = ": "),
+    span.dd = list(),
+
+    # This means that list elements have a margin, if they are nested
+    "ul ul li" = list("margin-left" = 2),
+    "ul ol li" = list("margin-left" = 2),
+    "ul dl li" = list("margin-left" = 2),
+    "ol ul li" = list("margin-left" = 2),
+    "ol ol li" = list("margin-left" = 2),
+    "ol dl li" = list("margin-left" = 2),
+    "ol ul li" = list("margin-left" = 2),
+    "ol ol li" = list("margin-left" = 2),
+    "ol dl li" = list("margin-left" = 2),
 
     blockquote = list("padding-left" = 4L, "padding-right" = 10L,
                       "font-style" = "italic", "margin-top" = 1L,
@@ -132,6 +197,7 @@ builtin_theme <- function(dark = getOption("cli_theme_dark", "auto")) {
     span.strong = list("font-weight" = "bold"),
     span.code = theme_code_tick(dark),
 
+    span.q   = list(fmt = quote_weird_name2),
     span.pkg = list(color = "blue"),
     span.fn = theme_function(dark),
     span.fun = theme_function(dark),
@@ -144,21 +210,74 @@ builtin_theme <- function(dark = getOption("cli_theme_dark", "auto")) {
     span.url = list(before = "<", after = ">", color = "blue",
                     "font-style" = "italic"),
     span.var = theme_code_tick(dark),
+    span.col = theme_code_tick(dark),
+    span.str = list(fmt = encode_string),
     span.envvar = theme_code_tick(dark),
     span.val = list(
       transform = function(x, ...) cli_format(x, ...),
       color = "blue"
     ),
-    span.field = list(color = "green")
+    span.field = list(color = "green"),
+    span.cls = list(collapse = "/", color = "blue", before = "<", after = ">"),
+    "span.progress-bar" = list(
+      transform = theme_progress_bar,
+      color = "green"
+    ),
+    span.or = list(vec_sep2 = " or ", vec_last = ", or "),
+    span.timestamp = list(before = "[", after = "]", color = "grey")
   )
 }
 
-quote_weird_name <- function(x) {
+encode_string <- function(x) {
+  encodeString(x, quote = "\"")
+}
+
+quote_weird_name0 <- function(x) {
+  x <- gsub(" ", "\u00a0", x)
   x2 <- ansi_strip(x)
-  if (!is_alnum(first_character(x2)) || !is_alnum(last_character(x2))) {
-    x <- paste0("'", x, "'")
+
+  fc <- first_character(x2)
+  sc <- second_character(x2)
+  lc <- last_character(x2)
+
+  wfst <- !is_alnum(fc, ok = "~") || (fc == "~" && !is_alnum(sc))
+  wlst <- !is_alnum(lc)
+
+  if (wfst || wlst) {
+    lsp <- leading_space(x2)
+    tsp <- trailing_space(x2)
+    if (nzchar(lsp)) {
+      x <- paste0(
+        bg_blue(lsp),
+        ansi_substr(x, nchar(lsp) + 1, ansi_nchar(x))
+      )
+    }
+    if (nzchar(tsp)) {
+      x <- paste0(
+        ansi_substr(x, 1, ansi_nchar(x) - nchar(tsp)),
+        bg_blue(tsp)
+      )
+    }
   }
-  x
+
+  list(x, wfst || wlst)
+}
+
+quote_weird_name <- function(x) {
+  x2 <- quote_weird_name0(x)
+  if (x2[[2]] || num_ansi_colors() == 1) {
+    x2[[1]] <- paste0("'", x2[[1]], "'")
+  }
+  x2[[1]]
+}
+
+quote_weird_name2 <- function(x) {
+  x2 <- quote_weird_name0(x)
+  paste0("\"", x2[[1]], "\"")
+}
+
+theme_progress_bar <- function(x, app, style) {
+  make_progress_bar(x$current / x$total, style = style)
 }
 
 detect_dark_theme <- function(dark) {
@@ -168,6 +287,8 @@ detect_dark_theme <- function(dark) {
         rstudioapi::getThemeInfo()$dark
       } else if (is_iterm()) {
         is_iterm_dark()
+      } else if (is_emacs()) {
+        Sys.getenv("ESS_BACKGROUND_MODE", "light") == "dark"
       } else {
         FALSE
       }
@@ -178,26 +299,42 @@ detect_dark_theme <- function(dark) {
 }
 
 theme_code <- function(dark) {
-  if (dark) {
-    list("background-color" = "#232323", color = "#d0d0d0")
-  } else{
-    list("background-color" = "#e8e8e8", color = "#202020")
-  }
+  list()
+}
+
+tick_formatter <- function(x) {
+  tt <- grepl("`", x, fixed = TRUE) + 1L
+  t1 <- c("`", "`` ")[tt]
+  t2 <- c("`", " ``")[tt]
+  paste0(t1, x, t2)
+}
+
+tick_formatter_fun <- function(x) {
+  tt <- grepl("`", x, fixed = TRUE) + 1L
+  t1 <- c("`", "`` ")[tt]
+  t2 <- c("()`", " ()``")[tt]
+  paste0(t1, x, t2)
 }
 
 theme_code_tick <- function(dark) {
-  modifyList(theme_code(dark), list(before = "`", after = "`"))
+  utils::modifyList(
+    theme_code(dark),
+    list(transform = tick_formatter)
+  )
 }
 
 theme_function <- function(dark) {
-  modifyList(theme_code(dark), list(before = "`", after = "()`"))
+  utils::modifyList(
+    theme_code(dark),
+    list(transform = tick_formatter_fun)
+  )
 }
 
 format_r_code <- function(dark) {
   function(x) {
     x <- ansi_strip(x)
     lines <- unlist(strsplit(x, "\n", fixed = TRUE))
-    tryCatch(prettycode::highlight(lines), error = function(x) lines)
+    code_highlight(lines)
   }
 }
 
@@ -233,6 +370,13 @@ create_formatter <- function(x) {
   if (!is_bold && !is_italic && !is_underline && !is_color
       && !is_bg_color) return(x)
 
+  if (is_color && is.null(x[["color"]])) {
+    x[["color"]] <- "none"
+  }
+  if (is_bg_color && is.null(x[["background-color"]])) {
+    x[["background-color"]] <- "none"
+  }
+
   fmt <- c(
     if (is_bold) list(style_bold),
     if (is_italic) list(style_italic),
@@ -253,24 +397,22 @@ create_formatter <- function(x) {
   x
 }
 
-#' @importFrom utils modifyList
-
 merge_embedded_styles <- function(old, new) {
-  # before and after is not inherited,
+  # before and after is not inherited, fmt is not inherited, either
   # side margins are additive, class mappings are merged
   # rest is updated, counter is reset
-  old$before <- old$after <- NULL
+  old$before <- old$after <- old$fmt <- NULL
 
   top <- new$`margin-top` %||% 0L
   bottom <- new$`margin-bottom` %||% 0L
   left <- (old$`margin-left` %||% 0L) + (new$`margin-left` %||% 0L)
   right <- (old$`margin-right` %||% 0L) + (new$`margin-right` %||% 0L)
 
-  map <- modifyList(old$`class-map` %||% list(), new$`class-map` %||% list())
+  map <- utils::modifyList(old$`class-map` %||% list(), new$`class-map` %||% list())
 
   start <- new$start %||% 1L
 
-  mrg <- modifyList(old, new)
+  mrg <- utils::modifyList(old, new)
   mrg[c("margin-top", "margin-bottom", "margin-left", "margin-right",
         "start", "class-map")] <- list(top, bottom, left, right, start, map)
 
@@ -287,7 +429,7 @@ merge_embedded_styles <- function(old, new) {
 #' * Type selectors, e.g. `input` selects all `<input>` elements.
 #' * Class selectors, e.g. `.index` selects any element that has a class
 #'   of "index".
-#' * ID selector. `#toc` will match the element that has the ID "toc".
+#' * ID selector. `#toc` will match the element that has the ID `"toc"`.
 #'
 #' Combinators:
 #'
@@ -336,7 +478,7 @@ parse_selector_node <- function(x) {
 #' * The id of the container is missing or unique.
 #' * The tag of the container is unique.
 #' * If the selector specifies an id, it matches the id of the container.
-#' * If the selector specifies a tag, it matxhes the tag of the container.
+#' * If the selector specifies a tag, it matches the tag of the container.
 #' * If the selector specifies class names, the container has all these
 #'   classes.
 #'
@@ -364,8 +506,18 @@ match_selector_node <- function(node, cnt) {
 match_selector <- function(sels, cnts) {
   sptr <- length(sels)
   cptr <- length(cnts)
+
+  # Last selector must match the last container
+  if (sptr == 0 || sptr > cptr) return(FALSE)
+  match <- match_selector_node(sels[[sptr]], cnts[[cptr]])
+  if (!match) return (FALSE)
+
+  # Plus the rest should match somehow
+  sptr <- sptr - 1L
+  cptr <- cptr - 1L
   while (sptr != 0L && sptr <= cptr) {
-    if (match_selector_node(sels[[sptr]], cnts[[cptr]])) {
+    match <- match_selector_node(sels[[sptr]], cnts[[cptr]])
+    if (match) {
       sptr <- sptr - 1L
       cptr <- cptr - 1L
     } else {
